<?php

namespace FiloBlu\Flow\Model\Actions\From;

/*
    The value on input can be:
    - a single number (58)
    - a single string (parka)
    - a path of strings (woman/jacket/parka) <-- search case insensitive based on the category admin value name
    - a path of numbers (2/5/9)
    - multiple single values (2||9) (woman||parka) (woman/jacket/parka||woman/jacket/coats) <-- multi values must be separated by double pipes '||'

    Backend configuration:
    - enable/disable import
    - enable/disable assign product to full category path
    - enable/disable assign categories only on configurables
    - enable/disable merge the new categories on the old ones

    Additional info:
    - Empty input value will be skipped (no categories will be deleted)
    - 'path of strings' need to start with the category root (Default category/SS18/Woman/Outerwear/PARKA|TRENCH)
    - 'paths' need to exist (after conversion in ids) inside the table 'catalog_category_entity'
*/

class Addtocategorynew {


    CONST ENTITY_TYPE_CODE = 'catalog_product';

    // The untouchable master root category (level 0)
    protected $_master_root_category_id = '1';

    protected $_categories_textstrpath_to_id = array();
    protected $_categories_textidpath_to_id = array();
    protected $_categories_full_ids_path_array_from_id = array();

    // All the root categories (level 1)
    protected $_root_category_id = array();

    protected $_logger;
    protected $_namesPaths = null;
    protected $_helper_product;
    protected $_configurable_type;
    protected $_category_factory;
    protected $_product_factory;


    public function __construct(
        \Magento\Framework\Model\Context $context,
        \Magento\Framework\Registry $registry,
        \FiloBlu\Flow\Helper\Product $helper_product,
        \FiloBlu\Flow\Helper\LoggerProvider $loggerProvider,
        \Magento\Catalog\Model\CategoryFactory $categoryFactory,
        \Magento\Catalog\Model\ProductFactory $productFactory,
        \Magento\ConfigurableProduct\Model\ResourceModel\Product\Type\Configurable $configurableProduct,
        array $data = []
    ) {
        $this->_logger = $loggerProvider->getLogger();
        $this->_helper_product = $helper_product;
        $this->_category_factory = $categoryFactory;
        $this->_product_factory = $productFactory;
        $this->_configurable_type = $configurableProduct;
        $this->_initCategories();
    }

    public function process($product, $value, $storeId) {

        if (empty($value)) {
            return true;
        }

        if(empty($this->_categories_full_ids_path_array_from_id) || empty($this->_categories_textstrpath_to_id) || empty($this->_categories_textidpath_to_id) || empty($this->_root_category_id)){
            $this->_initCategories();
        }

        $value = trim($value);

        $originalValue = $value;

        $_category_tree = array();

        // gestione piu' categorie
        // $_new_cat_ids will be always an array
        $_new_cat_ids = explode( '||', $originalValue);

        // security
        if(count($_new_cat_ids) == 0){
            return false;
        }

        // $cat_id can be:
        // single number (58)
        // single string (parka)
        // path of strings (woman/jacket/parka)
        // path of number (2/5/9)
        $categories_ids_array_to_add = array();
        foreach($_new_cat_ids as $_cat_id) {

            // This can return also multiple categories
            // Depends on the imput and backend config
            // Will return always an array
            $category_ids = $this->findTheCategoryId($_cat_id);

            if(count($category_ids) > 0) {
                foreach ($category_ids as $single_category) {
                    $categories_ids_array_to_add[] = $single_category;
                }
            }

            if(!$category_ids || count($category_ids) == 0){
                $this->_logger->info("Category '{$_cat_id}'' does not exist - SKIPPED ");
            }
        }

        // Nothing to do
        if(count($categories_ids_array_to_add) == 0){
            return true;
        }

        $merged_category = false;

        // First of all I need to add the categories on the configurable
        // Because the categories on the configurables will be always setted
        if ($product->getTypeId() == \Magento\Catalog\Model\Product\Type::DEFAULT_TYPE) {
            $parentIds = $this->_configurable_type->getParentIdsByChild($product->getId());
            if (count($parentIds) <= 0){
                return true;
            }
            foreach ($parentIds as $parentId) {
                $parent_product = $this->_product_factory->create()->load($parentId);
                // Salvataggio delle categorie
                $merged_category = $this->getCalculatedCategoriesList($parent_product,$categories_ids_array_to_add);
                $parent_product->setCategoryIds($merged_category)->save();
            }
        }

        // If I already have the categories from parent then those must be added to the simple
        // Else I need to calculate them
        if(!$merged_category){
            $merged_category = $this->getCalculatedCategoriesList($product,$categories_ids_array_to_add);
        }

        // Then I will for sure insert the categories on the simple if the simple have no parents and visibility != 1
        $simple_already_done = false;
        if ($product->getData('visibility') != '1') {
            $product->setCategoryIds($merged_category);
            $product->save();
            $simple_already_done = true;
        }

        //Now that the parents have the categories do I need to set them also on the simple?
        $_only_configurable_category = \Magento\Framework\App\ObjectManager::getInstance()
            ->create('\FiloBlu\Flow\Helper\Product')->getOnlyConfigurableCategoryFlag();

        // If not is only on configurable and I've not parsed already this product
        if(!$_only_configurable_category && !$simple_already_done){
            // Then add the categories also on this simple!
            $product->setCategoryIds($merged_category);
            $product->save();
        } else {
            // Or remove them because this simple should not have categories!
            $product->setCategoryIds(array());
            $product->save();
        }

        // Did everything, returning true
        return true;
    }

    public function assignFullPath(){
        $assign_full_path = \Magento\Framework\App\ObjectManager::getInstance()
            ->create('\FiloBlu\Flow\Helper\Product')->assignFullTree();
        return $assign_full_path;
    }

    public function mergeCategories(){
        $merge = \Magento\Framework\App\ObjectManager::getInstance()
            ->create('\FiloBlu\Flow\Helper\Product')->mergeCategoriesFlag();
        return $merge;
    }

    public function assignRootCategories(){
        $merge = \Magento\Framework\App\ObjectManager::getInstance()
            ->create('\FiloBlu\Flow\Helper\Product')->assignRootCategories();
        return $merge;
    }

    // This function will get the $product categories and handle if those need to be merged with the new ones or not
    // Returning the array of categories to be saved
    private function getCalculatedCategoriesList($product,$categories_ids_array_to_add){
        $_category_update = $this->mergeCategories();
        $merged_category = array();
        $old_categories = $product->getCategoryIds();
        if ($_category_update && count($old_categories)) {
            $merged_category = array_merge($old_categories, $categories_ids_array_to_add);
        } else {
            $merged_category = $categories_ids_array_to_add;
        }
        $merged_category = array_unique($merged_category);
        $merged_category = array_diff($merged_category, array($this->_master_root_category_id)); // Tolgo la master root category
        if(!$this->assignRootCategories() && count($this->_root_category_id) > 0){
            $merged_category = array_diff($merged_category, $this->_root_category_id); // Tolgo le root categories
        }
        return $merged_category;
    }


    /*
    $anything can be:
    > single number (9)
    > single string (parka)
    > path of strings (default category/woman/jacket/parka)
    > path of numbers (2/5/9)
    */
    public function findTheCategoryId($anything){

        $anything = trim(strtolower($anything));

        // single number (58)
        if (is_numeric($anything) && $anything > 0) {
            return array($anything);
        }

        $_categories_textstrpath_to_id = $this->_categories_textstrpath_to_id;
        $_categories_textidpath_to_id = $this->_categories_textidpath_to_id;
        $_categories_full_ids_path_array_from_id = $this->_categories_full_ids_path_array_from_id;

        // single string (parka)
        // path of strings (woman/jacket/parka)
        if(isset($_categories_textstrpath_to_id[$anything])){
            if($this->assignFullPath()){
                return $_categories_full_ids_path_array_from_id[$_categories_textstrpath_to_id[$anything]];
            } else {
                return array($_categories_textstrpath_to_id[$anything]);
            }
        }

        // path of numbers (2/5/9)
        if(isset($_categories_textidpath_to_id[$anything])){
            if($this->assignFullPath()){
                return $_categories_full_ids_path_array_from_id[$_categories_textidpath_to_id[$anything]];
            } else {
                return array($_categories_textidpath_to_id[$anything]);
            }
        }

        // not found
        return false;
    }

    protected function _initCategories()
    {
        $collection = \Magento\Framework\App\ObjectManager::getInstance()->create('\Magento\Catalog\Model\ResourceModel\Category\Collection');
        $collection->setStore(0);
        $collection->addNameToResult();

        foreach ($collection as $category) {
            $structure = explode('/', $category->getPath());

            // Saving the root categories ids
            if($category->getLevel() == 1){
                $this->_root_category_id[] = $category->getId();
            }

            // Removing the category ID 1, I don't need that
            if (($key = array_search('1', $structure)) !== false) {
                unset($structure[$key]);
            }

            // Generating path without the first master root category
            $path_no_master_root = implode("/",$structure);

            // Removing empty arrays, I don't need those
            if(count($structure) == 0){
                continue;
            }

            $path_with_names = array();
            foreach ($structure as $single_category_id) {
                $path_with_names[] = strtolower($collection->getItemById($single_category_id)->getName());
            }
            $full_string = implode("/",$path_with_names);

            $leaf_id = end($structure);

            $this->_categories_textstrpath_to_id[$full_string] = $leaf_id;
            $this->_categories_textidpath_to_id[$path_no_master_root] = $leaf_id;
            $this->_categories_full_ids_path_array_from_id[$leaf_id] = $structure;
        }

        return $this;
    }

}